@@ -34,6 +34,9 @@
 #'   bound to the symbol (the expressions are thus evaluated only
 #'   once, if at all).
 #'
+#' - `%<~%` is a shortcut for `env_bind_lazy()`. It works like `<-`
+#'   but the RHS is evaluated lazily.
+#'
 #'
 #' @section Side effects:
 #'
@@ -45,21 +48,13 @@
 #' Like other side-effecty functions like `par()` and `options()`,
 #' `env_bind()` and variants return the old values invisibly.
 #'
-#'
-#' @section Life cycle:
-#'
-#' Passing an environment wrapper like a formula or a function instead
-#' of an environment is soft-deprecated as of rlang 0.3.0. This
-#' internal genericity was causing confusion (see issue #427). You
-#' should now extract the environment separately before calling these
-#' functions.
-#'
 #' @param .env An environment.
 #' @param ... <[dynamic][dyn-dots]> Named objects (`env_bind()`),
 #'   expressions `env_bind_lazy()`, or functions (`env_bind_active()`).
 #'   Use [zap()] to remove bindings.
 #' @return The input object `.env`, with its associated environment
 #'   modified in place, invisibly.
+#' @seealso [env_poke()] for binding a single element.
 #' @export
 #' @examples
 #' # env_bind() is a programmatic way of assigning values to symbols
@@ -107,53 +102,27 @@
 #' # old values back:
 #' env_bind(my_env, !!!old)
 env_bind <- function(.env, ...) {
-  env_bind_impl(.env, list3(...), "env_bind()", bind = TRUE)
+  check_environment(.env)
+  invisible(.Call(
+    ffi_env_bind,
+    env = .env,
+    values = list3(...),
+    needs_old = TRUE,
+    bind_type = "value",
+    eval_env = NULL
+  ))
 }
-env_bind_impl <- function(env, data, fn, bind = FALSE, binder = NULL) {
-  if (!is_vector(data)) {
-    type <- friendly_type_of(data)
-    abort(sprintf("`data` must be a vector not a %s", type))
-  }
-  if (length(data)) {
-    nms <- names(data)
-    if (is_null(nms) || any(nms_are_invalid(nms))) {
-      abort("Can't bind data because all arguments must be named")
-    }
-    if (any(duplicated(nms))) {
-      abort("Can't bind data because some arguments have the same name")
-    }
-  }
 
-  env <- get_env_retired(env, fn)
-  data <- vec_coerce(data, "list")
-  nms <- names2(data)
-
-  if (bind) {
-    old <- new_list(length(nms), nms)
-    overwritten <- env_has(env, nms)
-    old[overwritten] <- env_get_list(env, nms[overwritten])
-    old[!overwritten] <- list(zap())
-
-    # We don't allow duplicates so we can remove bindings separately
-    zapped <- map_lgl(data, is_zap) & overwritten
-    rm(list = nms[zapped], envir = env)
-    data <- data[!zapped]
-    nms <- names(data) %||% chr()
-  }
-
-  if (is_null(binder)) {
-    .Call(rlang_env_bind_list, env, nms, data);
-  } else {
-    for (i in seq_along(data)) {
-      binder(env, nms[[i]], data[[i]])
-    }
-  }
-
-  if (bind) {
-    invisible(old)
-  } else {
-    invisible(env)
-  }
+# Doesn't return list of old bindings for efficiency
+env_bind0 <- function(.env, values) {
+  invisible(.Call(
+    ffi_env_bind,
+    env = .env,
+    values = values,
+    needs_old = FALSE,
+    bind_type = "value",
+    eval_env = NULL
+  ))
 }
 
 #' @rdname env_bind
@@ -199,19 +168,15 @@ env_bind_impl <- function(env, data, fn, bind = FALSE, binder = NULL) {
 #' env_bind_lazy(env, name = !!quo)
 #' env$name
 env_bind_lazy <- function(.env, ..., .eval_env = caller_env()) {
-  exprs <- exprs(...)
-  exprs <- map_if(exprs, is_quosure, function(quo) call2(as_function(quo)))
-
-  binder <- function(env, nm, value) {
-    do.call("delayedAssign", list(
-      x = nm,
-      value = value,
-      eval.env = .eval_env,
-      assign.env = env
-    ))
-  }
-
-  env_bind_impl(.env, exprs, "env_bind_lazy()", TRUE, binder)
+  check_environment(.env)
+  invisible(.Call(
+    ffi_env_bind,
+    env = .env,
+    values = exprs(...),
+    needs_old = TRUE,
+    bind_type = "lazy",
+    eval_env = .eval_env
+  ))
 }
 #' @rdname env_bind
 #' @export
@@ -239,19 +204,32 @@ env_bind_lazy <- function(.env, ..., .eval_env = caller_env()) {
 #' env$foo
 #' env$foo
 env_bind_active <- function(.env, ...) {
-  fns <- map_if(list3(...), negate(is_zap), as_function)
-
-  existing <- env_names(.env)
-  binder <- function(env, nm, value) {
-    # makeActiveBinding() fails if there is already a regular binding
-    if (nm %in% existing) {
-      env_unbind(env, nm)
-    }
-    makeActiveBinding(nm, value, env)
-  }
-
-  env_bind_impl(.env, fns, "env_bind_active()", TRUE, binder)
+  check_environment(.env)
+  invisible(.Call(
+    ffi_env_bind,
+    env = .env,
+    values = list3(...),
+    needs_old = TRUE,
+    bind_type = "active",
+    eval_env = caller_env()
+  ))
 }
+#' @rdname env_bind
+#' @param lhs The variable name to which `rhs` will be lazily assigned.
+#' @param rhs An expression lazily evaluated and assigned to `lhs`.
+#' @export
+`%<~%` <- function(lhs, rhs) {
+  env <- caller_env()
+  inject(
+    base::delayedAssign(
+      as_string(substitute(lhs)),
+      !!substitute(rhs),
+      eval.env = env,
+      assign.env = env
+    )
+  )
+}
+
 
 #' Temporarily change bindings of an environment
 #'
@@ -283,15 +261,11 @@ env_bind_active <- function(.env, ...) {
 #' with_bindings(paste(foo, bar), foo = "rebinded")
 #' paste(foo, bar)
 local_bindings <- function(..., .env = .frame, .frame = caller_env()) {
-  env <- get_env_retired(.env, "local_bindings()")
+  check_environment(.env)
+  check_environment(.frame)
 
-  bindings <- list2(...)
-  if (!length(bindings)) {
-    return(list())
-  }
-
-  old <- env_bind_impl(env, bindings, "local_bindings()", bind = TRUE)
-  local_exit(frame = .frame, !!call2(env_bind_impl, env, old, bind = TRUE))
+  old <- env_bind(.env, ...)
+  defer(env_bind0(.env, old), envir = .frame)
 
   invisible(old)
 }
@@ -299,46 +273,11 @@ local_bindings <- function(..., .env = .frame, .frame = caller_env()) {
 #' @param .expr An expression to evaluate with temporary bindings.
 #' @export
 with_bindings <- function(.expr, ..., .env = caller_env()) {
-  env <- get_env_retired(.env, "with_bindings()")
+  check_environment(.env)
   local_bindings(..., .env = .env)
   .expr
 }
 
-#' Mask bindings by defining symbols deeper in a scope
-#'
-#' `env_bury()` is like [env_bind()] but it creates the bindings in a
-#' new child environment. This makes sure the new bindings have
-#' precedence over old ones, without altering existing environments.
-#' Unlike `env_bind()`, this function does not have side effects and
-#' returns a new environment (or object wrapping that environment).
-#'
-#' @inheritParams env_bind
-#' @return A copy of `.env` enclosing the new environment containing
-#'   bindings to `...` arguments.
-#' @seealso [env_bind()], [env_unbind()]
-#'
-#' @keywords internal
-#' @export
-#' @examples
-#' orig_env <- env(a = 10)
-#' fn <- set_env(function() a, orig_env)
-#'
-#' # fn() currently sees `a` as the value `10`:
-#' fn()
-#'
-#' # env_bury() will bury the current scope of fn() behind a new
-#' # environment:
-#' fn <- env_bury(fn, a = 1000)
-#' fn()
-#'
-#' # Even though the symbol `a` is still defined deeper in the scope:
-#' orig_env$a
-env_bury <- function(.env, ...) {
-  env_ <- get_env(.env)
-  env_ <- child_env(env_, ...)
-  set_env(.env, env_)
-}
-
 #' Remove bindings from an environment
 #'
 #' `env_unbind()` is the complement of [env_bind()]. Like `env_has()`,
@@ -346,38 +285,30 @@ env_bury <- function(.env, ...) {
 #' `inherit` to `TRUE` to track down bindings in parent environments.
 #'
 #' @inheritParams get_env
-#' @param nms A character vector containing the names of the bindings
-#'   to remove.
+#' @param nms A character vector of binding names to remove.
 #' @param inherit Whether to look for bindings in the parent
 #'   environments.
 #' @return The input object `env` with its associated environment
 #'   modified in place, invisibly.
 #' @export
 #' @examples
-#' data <- set_names(as.list(letters), letters)
-#' env_bind(environment(), !!! data)
-#' env_has(environment(), letters)
+#' env <- env(foo = 1, bar = 2)
+#' env_has(env, c("foo", "bar"))
 #'
-#' # env_unbind() removes bindings:
-#' env_unbind(environment(), letters)
-#' env_has(environment(), letters)
+#' # Remove bindings with `env_unbind()`
+#' env_unbind(env, c("foo", "bar"))
+#' env_has(env, c("foo", "bar"))
 #'
 #' # With inherit = TRUE, it removes bindings in parent environments
 #' # as well:
-#' parent <- child_env(NULL, foo = "a")
-#' env <- child_env(parent, foo = "b")
+#' parent <- env(empty_env(), foo = 1, bar = 2)
+#' env <- env(parent, foo = "b")
+#'
 #' env_unbind(env, "foo", inherit = TRUE)
-#' env_has(env, "foo", inherit = TRUE)
+#' env_has(env, c("foo", "bar"))
+#' env_has(env, c("foo", "bar"), inherit = TRUE)
 env_unbind <- function(env = caller_env(), nms, inherit = FALSE) {
-  if (inherit) {
-    while (any(exist <- env_has(env, nms, inherit = TRUE))) {
-      rm(list = nms[exist], envir = env, inherits = TRUE)
-    }
-  } else {
-    exist <- env_has(env, nms, inherit = FALSE)
-    rm(list = nms[exist], envir = env)
-  }
-
+  .Call(ffi_env_unbind, env, nms, inherit)
   invisible(env)
 }
 
@@ -389,6 +320,8 @@ env_unbind <- function(env = caller_env(), nms, inherit = FALSE) {
 #' any of its parents (with `inherit = TRUE`).
 #'
 #' @inheritParams env_unbind
+#' @param nms A character vector of binding names for which to check
+#'   existence.
 #' @return A named logical vector as long as `nms`.
 #' @export
 #' @examples
@@ -399,9 +332,8 @@ env_unbind <- function(env = caller_env(), nms, inherit = FALSE) {
 #' env_has(env, "foo")
 #' env_has(env, "foo", inherit = TRUE)
 env_has <- function(env = caller_env(), nms, inherit = FALSE) {
-  env <- get_env_retired(env, "env_has()")
-  nms <- set_names(nms)
-  map_lgl(nms, exists, envir = env, inherits = inherit)
+  check_environment(env)
+  .Call(ffi_env_has, env, nms, inherit)
 }
 
 #' Get an object in an environment
@@ -416,7 +348,12 @@ env_has <- function(env = caller_env(), nms, inherit = FALSE) {
 #' @param nm,nms Names of bindings. `nm` must be a single string.
 #' @param default A default value in case there is no binding for `nm`
 #'   in `env`.
+#' @param last Last environment inspected when `inherit` is `TRUE`.
+#'   Can be useful in conjunction with [base::topenv()].
 #' @return An object if it exists. Otherwise, throws an error.
+#'
+#' @seealso [env_cache()] for a variant of `env_get()` designed to
+#'   cache a value in an environment.
 #' @export
 #' @examples
 #' parent <- child_env(NULL, foo = "foo")
@@ -430,43 +367,53 @@ env_has <- function(env = caller_env(), nms, inherit = FALSE) {
 #'
 #' # You can also avoid an error by supplying a default value:
 #' env_get(env, "foo", default = "FOO")
-env_get <- function(env = caller_env(), nm, default, inherit = FALSE) {
-  env <- get_env_retired(env, "env_get()")
-  if (!missing(default)) {
-    exists <- env_has(env, nm, inherit = inherit)
-    if (!exists) {
-      return(default)
-    }
+env_get <- function(env = caller_env(),
+                    nm,
+                    default,
+                    inherit = FALSE,
+                    last = empty_env()) {
+  check_environment(env)
+  check_environment(last)
+
+  if (missing(default)) {
+    default %<~% stop_env_get_missing(nm)
   }
 
-  # FIXME: The `inherit` case fails with missing arguments
-  if (inherit) {
-    return(get(nm, envir = env, inherits = TRUE))
-  }
-
-  .Call(rlang_env_get, env, nm)
+  .Call(
+    ffi_env_get,
+    env = env,
+    nm = nm,
+    inherit = inherit,
+    last = last,
+    closure_env = environment()
+  )
 }
 #' @rdname env_get
 #' @export
-env_get_list <- function(env = caller_env(), nms, default, inherit = FALSE) {
-  env <- get_env_retired(env, "env_get_list()")
+env_get_list <- function(env = caller_env(),
+                         nms,
+                         default,
+                         inherit = FALSE,
+                         last = empty_env()) {
+  check_environment(env)
+  check_environment(last)
+  .Call(
+    ffi_env_get_list,
+    env = env,
+    nms = nms,
+    inherit = inherit,
+    last = last,
+    closure_env = environment()
+  )
+}
 
-  nms <- set_names(nms)
-
-  # FIXME R 3.1: missingness of `default` is not passed through dots on 3.1
-  if (missing(default)) {
-    map(nms, env_get, env = env, inherit = inherit)
-  } else {
-    map(nms, env_get, env = env, inherit = inherit, default = default)
-  }
+stop_env_get_missing <- function(nm) {
+  msg <- sprintf("Can't find %s in environment.", format_arg(nm))
+  abort(msg, call = caller_env())
 }
 
 #' Poke an object in an environment
 #'
-#' @description
-#'
-#' \Sexpr[results=rd, stage=render]{rlang:::lifecycle("experimental")}
-#'
 #' `env_poke()` will assign or reassign a binding in `env` if `create`
 #' is `TRUE`. If `create` is `FALSE` and a binding does not already
 #' exists, an error is issued.
@@ -488,13 +435,6 @@ env_get_list <- function(env = caller_env(), nms, default, inherit = FALSE) {
 #' environment when no existing binding is found in the parents.
 #'
 #'
-#' @section Life cycle:
-#'
-#' `env_poke()` is experimental. We are still experimenting with
-#' reducing the number of redundant functions by using quasiquotation.
-#' It is possible `env_poke()` will be deprecated in favour of
-#' `env_bind()` and name-unquoting with `:=`.
-#'
 #' @inheritParams env_get
 #' @param value The value for a new binding.
 #' @param create Whether to create a binding if it does not already
@@ -502,40 +442,62 @@ env_get_list <- function(env = caller_env(), nms, default, inherit = FALSE) {
 #' @return The old value of `nm` or a [zap sentinel][zap] if the
 #'   binding did not exist yet.
 #'
-#' @keywords internal
+#' @seealso [env_bind()] for binding multiple elements. [env_cache()]
+#'   for a variant of `env_poke()` designed to cache values.
 #' @export
-env_poke <- function(env = caller_env(), nm, value,
-                     inherit = FALSE, create = !inherit) {
-  stopifnot(is_string(nm))
-  env_ <- get_env_retired(env, "env_poke()")
-  old <- env_get(env_, nm, inherit = inherit, default = zap())
-
-  if (inherit) {
-    scope_poke(env_, nm, value, create)
-  } else if (create || env_has(env_, nm)) {
-    assign(nm, value, envir = env_)
-  } else {
-    abort(paste0("Can't find existing binding in `env` for \"", nm, "\""))
-  }
-
-  invisible(maybe_missing(old))
+env_poke <- function(env = caller_env(),
+                     nm,
+                     value,
+                     inherit = FALSE,
+                     create = !inherit) {
+  check_environment(env)
+  invisible(.Call(
+    ffi_env_poke,
+    env = env,
+    nm = nm,
+    values = value,
+    inherit = inherit,
+    create = create
+  ))
 }
-scope_poke <- function(env, nm, value, create) {
-  cur <- env
 
-  while (!env_has(cur, nm) && !is_empty_env(cur)) {
-    cur <- env_parent(cur)
+#' Cache a value in an environment
+#'
+#' @description
+#' `env_cache()` is a wrapper around [env_get()] and [env_poke()]
+#' designed to retrieve a cached value from `env`.
+#'
+#' - If the `nm` binding exists, it returns its value.
+#' - Otherwise, it stores the default value in `env` and returns that.
+#'
+#' @inheritParams env_get
+#' @param default The default value to store in `env` if `nm` does not
+#'   exist yet.
+#' @return Either the value of `nm` or `default` if it did not exist
+#'   yet.
+#'
+#' @examples
+#' e <- env(a = "foo")
+#'
+#' # Returns existing binding
+#' env_cache(e, "a", "default")
+#'
+#' # Creates a `b` binding and returns its default value
+#' env_cache(e, "b", "default")
+#'
+#' # Now `b` is defined
+#' e$b
+#' @export
+env_cache <- function(env, nm, default) {
+  check_required(default)
+  check_string(nm)
+
+  if (env_has(env, nm)) {
+    env_get(env, nm)
+  } else {
+    env_poke(env, nm, default)
+    default
   }
-
-  if (is_empty_env(cur)) {
-    if (!create) {
-      abort(paste0("Can't find existing binding in `env` for \"", nm, "\""))
-    }
-    cur <- env
-  }
-
-  assign(nm, value, envir = cur)
-  env
 }
 
 #' Names and numbers of symbols bound in an environment
@@ -575,17 +537,15 @@ scope_poke <- function(env, nm, value, create) {
 #' env <- env(a = 1, b = 2)
 #' env_names(env)
 env_names <- function(env) {
-  env <- get_env_retired(env, "env_names()")
+  check_environment(env)
   nms <- names(env)
-  .Call(rlang_unescape_character, nms)
+  .Call(ffi_unescape_character, nms)
 }
 
 #' @rdname env_names
 #' @export
 env_length <- function(env) {
-  if (!is_environment(env)) {
-    abort("`env` must be an environment")
-  }
+  check_environment(env)
   length(env)
 }
 
@@ -593,7 +553,7 @@ env_length <- function(env) {
 #'
 #' @description
 #'
-#' \Sexpr[results=rd, stage=render]{rlang:::lifecycle("experimental")}
+#' `r lifecycle::badge("experimental")`
 #'
 #' Locked environment bindings trigger an error when an attempt is
 #' made to redefine the binding.
@@ -655,7 +615,7 @@ env_binding_are_locked <- function(env, nms = NULL) {
 
 #' What kind of environment binding?
 #'
-#' \Sexpr[results=rd, stage=render]{rlang:::lifecycle("experimental")}
+#' `r lifecycle::badge("experimental")`
 #'
 #' @inheritParams env_binding_lock
 #'
@@ -670,11 +630,13 @@ env_binding_are_active <- function(env, nms = NULL) {
 env_binding_are_lazy <- function(env, nms = NULL) {
   env_binding_are_type(env, nms, 1L)
 }
-env_binding_are_type <- function(env, nms, type) {
-  if (!is_environment(env)) {
-    abort("`env` must be an environment.")
-  }
-  nms <- env_binding_validate_names(env, nms)
+env_binding_are_type <- function(env,
+                                 nms,
+                                 type,
+                                 error_call = caller_env()) {
+  check_environment(env, call = error_call)
+
+  nms <- env_binding_validate_names(env, nms, call = error_call)
   promise <- env_binding_types(env, nms)
 
   if (is_null(promise)) {
@@ -685,18 +647,20 @@ env_binding_are_type <- function(env, nms, type) {
   set_names(promise, nms)
 }
 
-env_binding_validate_names <- function(env, nms) {
+env_binding_validate_names <- function(env, nms, call = caller_env()) {
   if (is_null(nms)) {
     nms <- env_names(env)
   } else {
-    if (!is_character(nms)) {
-      abort("`nms` must be a character vector of names")
-    }
+    check_character(
+      nms,
+      what = "a character vector of names",
+      call = call
+    )
   }
   nms
 }
 env_binding_types <- function(env, nms = env_names(env)) {
-  .Call(rlang_env_binding_types, env, nms)
+  .Call(ffi_env_binding_types, env, nms)
 }
 
 env_binding_type_sum <- function(env, nms = NULL) {
