@@ -101,8 +101,8 @@ peek_option <- function(name) {
 #'   escape hatch is useful in unit tests or to manually turn on
 #'   interactive features in RMarkdown outputs.
 #'
-#' * Whether knitr, an RStudio notebook, or testthat is in progress
-#'   (in which case `is_interactive()` returns `FALSE`).
+#' * Whether knitr or testthat is in progress, in which case
+#'   `is_interactive()` returns `FALSE`.
 #'
 #' `with_interactive()` and `local_interactive()` set the global
 #' option conveniently.
@@ -113,7 +113,7 @@ is_interactive <- function() {
   if (!is_null(opt)) {
     if (!is_bool(opt)) {
       options(rlang_interactive = NULL)
-      abort("`rlang_interactive` must be a single `TRUE` of `FALSE`")
+      check_bool(opt, arg = "rlang_interactive")
     }
     return(opt)
   }
@@ -121,9 +121,6 @@ is_interactive <- function() {
   if (is_true(peek_option("knitr.in.progress"))) {
     return(FALSE)
   }
-  if (is_true(peek_option("rstudio.notebook.executing"))) {
-    return(FALSE)
-  }
   if (identical(Sys.getenv("TESTTHAT"), "true")) {
     return(FALSE)
   }
@@ -148,3 +145,19 @@ with_interactive <- function(expr, value = TRUE) {
   local_interactive(value)
   expr
 }
+
+report_in_progress <- function() {
+  if (is_true(peek_option("knitr.in.progress"))) {
+    return(TRUE)
+  }
+
+  if (is_true(peek_option("rstudio.notebook.executing"))) {
+    return(TRUE)
+  }
+
+  FALSE
+}
+
+peek_srcref <- function() {
+  .Call(ffi_peek_srcref)
+}
